;; Generated package description from el-search.el
(define-package "el-search" "1.2.3" "Expression based interactive search for emacs-lisp-mode" '((emacs "25") (stream "2.2.3")) :url "http://elpa.gnu.org/packages/el-search.html" :keywords '("lisp"))
